import logging

from PyQt5 import Qt
from PyQt5.QtWidgets import *

import envi.qt.memory as e_mem_qt
import envi.qt.memcanvas as e_mem_canvas

import vqt.hotkeys as vq_hotkey

import vivisect.base as viv_base
import vivisect.renderers as viv_rend
import vivisect.qt.views as viv_q_views
import vivisect.qt.ctxmenu as viv_q_ctxmenu

from vqt.main import *
from vivisect.const import *

logger = logging.getLogger(__name__)


# FIXME HACK where do these really live?
qt_horizontal   = 1
qt_vertical     = 2

class VivCanvasBase(vq_hotkey.HotKeyMixin, e_mem_canvas.VQMemoryCanvas):

    def __init__(self, *args, **kwargs):

        e_mem_canvas.VQMemoryCanvas.__init__(self, *args, **kwargs)
        vq_hotkey.HotKeyMixin.__init__(self)

        self.vw = self.mem
        self._last_sname = None

        self.addHotKey('c', 'viv:make:code')
        self.addHotKey('f', 'viv:make:function')
        self.addHotKey('s', 'viv:make:string')
        self.addHotKey('p', 'viv:make:pointer')
        self.addHotKey('u', 'viv:make:unicode')
        self.addHotKey('n', 'viv:setname')
        self.addHotKey('g', 'viv:getlocation')
        self.addHotKey(';', 'viv:comment')
        self.addHotKey(':', 'viv:commentmulti')
        self.addHotKey('t', 'viv:make:struct')
        self.addHotKey('ctrl+t', 'viv:make:struct:again')
        self.addHotKey('ctrl+meta+t', 'viv:make:struct:multi')
        self.addHotKey('U', 'viv:undefine')
        self.addHotKey('ctrl+p', 'viv:preview:instr')
        self.addHotKey('B', 'viv:bookmark')
        self.addHotKey('ctrl+meta+J', 'viv:javascript')
        self.addHotKey('P', 'viv:make:ptr:array')

        self.addHotKey('ctrl+1', 'viv:make:number:one')
        self.addHotKey('ctrl+2', 'viv:make:number:two')
        self.addHotKey('ctrl+4', 'viv:make:number:four')
        self.addHotKey('ctrl+6', 'viv:make:number:sixteen')
        self.addHotKey('ctrl+8', 'viv:make:number:eight')

        self.addHotKey('down', 'viv:nav:nextva')
        self.addHotKey('up', 'viv:nav:prevva')
        self.addHotKey('pgdown', 'viv:nav:nextpage')
        self.addHotKey('pgup', 'viv:nav:prevpage')
        self.addHotKey('ctrl+down', 'viv:nav:nextundef')
        self.addHotKey('ctrl+up', 'viv:nav:prevundef')

        self.loadHotKeys(self.vw._viv_gui._vq_settings)

        # All extenders must implement vivColorMap
        vqtconnect(self.vivColorMap, 'viv:colormap')

    def event(self, evt):
        if evt.type() == Qt.QEvent.ChildAdded:
            evt.child().installEventFilter(self)
        elif evt.type() == Qt.QEvent.ChildRemoved:
            evt.child().removeEventFilter(self)
        return e_mem_canvas.VQMemoryCanvas.event(self, evt)

    def eventFilter(self, src, evt):
        if evt.type() == Qt.QEvent.KeyPress:
            return self.eatKeyPressEvent(evt)
        return False

    def vivColorMap(self, event, einfo):
        self._applyColorMap(einfo)

    def _applyColorMap(self, cmap):

        page = self.page()
        inner = ''
        for va, color in cmap.items():
            inner += '.envi-va-0x%.8x { color: #000000; background-color: %s }\n' % (va, color)
        inner = inner.replace('`', r'\`')
        js = 'var node = document.querySelector("#cmapstyle"); node.innerHTML = `%s`;' % inner
        page.runJavaScript(js)

    @vq_hotkey.hotkey('viv:nav:nextva')
    def _hotkey_nav_nextva(self):
        if self._canv_curva is None:
            return

        loc = self.vw.getLocation(self._canv_curva)
        if loc is None:
            loc = (self._canv_curva, 1, None, None)

        nextva = loc[0] + loc[1]
        self._selectVa(nextva)

    @vq_hotkey.hotkey('viv:nav:prevva')
    def _hotkey_nav_prevva(self):
        if self._canv_curva is None:
            return

        loc = self.vw.getPrevLocation(self._canv_curva)
        if loc is None:
            loc = (self._canv_curva - 1, 1, None, None)

        self._selectVa(loc[0])

    @vq_hotkey.hotkey('viv:nav:nextpage')
    def _hotkey_nav_nextpage(self):
        if self._canv_curva is None:
            return

        self.scrollEvent()
        return False

    @vq_hotkey.hotkey('viv:nav:prevpage')
    def _hotkey_nav_prevpage(self):
        if self._canv_curva is None:
            return

        self.scrollEvent()
        return False

    @vq_hotkey.hotkey('viv:nav:nextundef')
    def _hotkey_nav_nextundef(self):
        if self._canv_curva is None:
            return

        vw = self.vw
        va = self._canv_curva
        loc = vw.getLocation(va)

        if loc is None:
            # find next defined location
            while loc is None and vw.isValidPointer(va):
                va += 1
                loc = vw.getLocation(va)
            va -= 1
            lastloc = (va, 1, 0, 0)
        else:
            # find next undefined location
            while loc is not None:
                va = loc[0]
                lastloc = loc
                loc = vw.getLocation(va + loc[1])

        # if we didn't fall off the map
        if vw.isValidPointer(va+lastloc[1]):
            va += lastloc[1]
        self._navExpression(hex(va))

    @vq_hotkey.hotkey('viv:nav:prevundef')
    def _hotkey_nav_prevundef(self):
        if self._canv_curva is None:
            return

        vw = self.vw
        va = self._canv_curva
        loc = vw.getLocation(va)

        if loc is None:
            # find previous defined location
            while loc is None and vw.isValidPointer(va):
                va -= 1
                loc = vw.getLocation(va)
            if loc is not None:
                va = loc[0]
        else:
            # find previous undefined location
            while loc is not None:
                va = loc[0]
                loc = vw.getLocation(va-1)

            # if we fell off the end of a map
            if vw.isValidPointer(va-1):
                va -= 1

        self._navExpression(hex(va))

    @vq_hotkey.hotkey('viv:make:code')
    def _hotkey_make_code(self):
        if self._canv_curva is not None:
            loctup = self.vw.getLocation(self._canv_curva)
            if loctup is not None:
                self.vw.vprint("Failed to make code (at 0x%x) where something already exists: %r " % (va, self.reprLocation(loctup)))
                return

            self.vw.makeCode(self._canv_curva)

    @vq_hotkey.hotkey('viv:make:function')
    def _hotkey_make_function(self):
        if self._canv_curva is not None:
            logger.debug('new function (manual): 0x%x', self._canv_curva)
            self.vw.makeFunction(self._canv_curva)

    @vq_hotkey.hotkey('viv:make:string')
    def _hotkey_make_string(self):
        if self._canv_curva is not None:
            loctup = self.vw.getLocation(self._canv_curva)
            if loctup is not None:
                self.vw.vprint("Failed to make string (at 0x%x) where something already exists: %r " % (va, self.reprLocation(loctup)))
                return

            self.vw.makeString(self._canv_curva)

    @vq_hotkey.hotkey('viv:make:pointer')
    def _hotkey_make_pointer(self):
        if self._canv_curva is not None:
            loctup = self.vw.getLocation(self._canv_curva)
            if loctup is not None:
                self.vw.vprint("Failed to make pointer (at 0x%x) where something already exists: %r " % (va, self.reprLocation(loctup)))
                return

            self.vw.makePointer(self._canv_curva)

    @vq_hotkey.hotkey('viv:make:ptr:array')
    def _hotkey_make_pointer_array(self):
        if self._canv_curva is not None:
            loctup = self.vw.getLocation(self._canv_curva)
            if loctup is not None:
                self.vw.vprint("Failed to make pointer array (at 0x%x) where something already exists: %r " % (va, self.reprLocation(loctup)))
                return

            self.vw.getVivGui().makePtrArray(self._canv_curva)

    @vq_hotkey.hotkey('viv:make:unicode')
    def _hotkey_make_unicode(self):
        if self._canv_curva is not None:
            loctup = self.vw.getLocation(self._canv_curva)
            if loctup is not None:
                self.vw.vprint("Failed to make unicode (at 0x%x) where something already exists: %r " % (va, self.reprLocation(loctup)))
                return

            self.vw.makeUnicode(self._canv_curva)

    @vq_hotkey.hotkey('viv:undefine')
    def _hotkey_undefine(self):
        if self._canv_curva is not None:
            self.vw.delLocation(self._canv_curva)

    @vq_hotkey.hotkey('viv:getlocation')
    def _hotkey_getlocation(self):
        if self._canv_curva is not None:
            self.vw.getVivGui().getLocation(self._canv_curva)

    @vq_hotkey.hotkey('viv:setname')
    def _hotkey_setname(self):
        if self._canv_curva is not None:
            self.vw.getVivGui().setName(self._canv_curva, self._canv_curtag, parent=self)

    @vq_hotkey.hotkey('viv:bookmark')
    def _hotkey_bookmark(self):
        if self._canv_curva is not None:
            self.vw.getVivGui().addBookmark(self._canv_curva, parent=self)

    @vq_hotkey.hotkey('viv:comment')
    def _hotkey_comment(self):
        if self._canv_curva is not None:
            self.vw.getVivGui().setVaComment(self._canv_curva, parent=self)

    @vq_hotkey.hotkey('viv:commentmulti')
    def _hotkey_commentmulti(self):
        if self._canv_curva is not None:
            self.vw.getVivGui().setVaMultilineComment(self._canv_curva, parent=self)

    @vq_hotkey.hotkey('viv:make:struct')
    def _hotkey_make_struct(self):
        if self._canv_curva is not None:
            sname = self.vw.getVivGui().makeStruct(self._canv_curva)
            if sname is not None:
                self._last_sname = sname

    @vq_hotkey.hotkey('viv:make:struct:again')
    def _hotkey_make_struct_again(self):
        if self._canv_curva is not None:
            if self._last_sname is not None:
                self.vw.makeStructure(self._canv_curva, self._last_sname)

    @vq_hotkey.hotkey('viv:make:struct:multi')
    def _hotkey_make_struct_multi(self, parent=None):
        if self._canv_curva is not None:
            if self._last_sname is not None:
                number, ok = QInputDialog.getText(parent, 'Make Multiple Consecutive Structs', 'Number of Structures')
                if ok:
                    curva = self._canv_curva
                    number = int(str(number), 0)
                    for count in range(number):
                        vs = self.vw.makeStructure(curva, self._last_sname)
                        curva += len(vs)

    @vq_hotkey.hotkey('viv:javascript')
    def _hotkey_dbg_runjavascript(self, parent=None):
        js, ok = QInputDialog.getText(parent, 'Run Javascript', 'code:')
        if ok:
            self.page().runJavaScript(js)

    def makeStructAgainMulti(self, va, parent=None):
        if parent is None:
            parent = self

        curcomment = self.vw.getComment(va)
        if curcomment is None:
            curcomment = ''

        comment, ok = QInputDialog.getText(parent, 'Enter...', 'Comment', text=curcomment)
        if ok:
            self.vw.setComment(va, str(comment))

    @vq_hotkey.hotkey('viv:make:number:one')
    def _hotkey_make_number_one(self):
        if self._canv_curva is not None:
            self.vw.makeNumber(self._canv_curva, 1)

    @vq_hotkey.hotkey('viv:make:number:two')
    def _hotkey_make_number_two(self):
        if self._canv_curva is not None:
            self.vw.makeNumber(self._canv_curva, 2)

    @vq_hotkey.hotkey('viv:make:number:four')
    def _hotkey_make_number_four(self):
        if self._canv_curva is not None:
            self.vw.makeNumber(self._canv_curva, 4)

    @vq_hotkey.hotkey('viv:make:number:eight')
    def _hotkey_make_number_eight(self):
        if self._canv_curva is not None:
            self.vw.makeNumber(self._canv_curva, 8)

    @vq_hotkey.hotkey('viv:make:number:sixteen')
    def _hotkey_make_number_sixteen(self):
        if self._canv_curva is not None:
            self.vw.makeNumber(self._canv_curva, 16)

    @vq_hotkey.hotkey('viv:preview:instr')
    def _hotkey_preview_instr(self):
        if self._canv_curva is not None:
            self.vw.previewCode(self._canv_curva)

    def getVaTag(self, va):
        loc = self.mem.getLocation(va)
        if loc is not None:
            va = loc[L_VA]
        return e_mem_canvas.VQMemoryCanvas.getVaTag(self, va)


class VQVivMemoryCanvas(VivCanvasBase):

    def _scrollEventCallback(self, data):
        '''
        Ugh. Yes. I know this sucks.
        But we have to do this because QtWebEngine does't natively let you get the max scroll size.
        You *have* to go through javascript to get those elements, and the only way to be sure of
        the function finishing (and being able to get a value outta js) is via this callback
        mechanism they set up.
        '''
        if not data:
            return
        smin = data[0]
        spos = data[1]
        smax = data[2]
        if not len(self._canv_rendvas):
            pass

        elif spos >= smax:
            lastva, lastsize = self._canv_rendvas[-1]
            mapva, mapsize, mperm, mfname = self.vw.getMemoryMap(lastva)
            sizeremain = (mapva + mapsize) - (lastva + lastsize)
            if sizeremain:
                self.renderMemoryAppend(min(sizeremain, 128))

        elif spos == smin:
            firstva, firstsize = self._canv_rendvas[0]
            mapva, mapsize, mperm, mfname = self.vw.getMemoryMap(firstva)
            sizeremain = firstva - mapva
            if sizeremain:
                self.renderMemoryPrepend(min(sizeremain, 128))

    def scrollEvent(self):
        page = self.page()
        page.runJavaScript('''
        var pcur = window.innerHeight + window.pageYOffset
        var scrollMaxY = Math.max(
            document.body.scrollHeight, document.documentElement.scrollHeight,
            document.body.offsetHeight, document.documentElement.offsetHeight,
            document.body.clientHeight, document.documentElement.clientHeight,
        );
        [window.innerHeight, pcur, scrollMaxY];
        ''', self._scrollEventCallback)

    def wheelEvent(self, event):
        self.scrollEvent()
        return e_mem_canvas.VQMemoryCanvas.wheelEvent(self, event)

    def _clearColorMap(self):
        page = self.page()
        page.runJavaScript('var node = document.querySelector("#cmapstyle"); node.innerHTML = "";')

    def _navExpression(self, expr):
        if self._canv_navcallback:
            self._canv_navcallback(expr)

    def initMemWindowMenu(self, va, tag, menu):
        nav = self.parent()  # our parent is always a VQVivMemoryWindow (nav target)
        viv_q_ctxmenu.buildContextMenu(self.vw, va=va, menu=menu, nav=nav, tag=tag)

    def _loc_helper(self, va):
        '''
        we assume we're being handed a valid va since renderMemory checks for valid MemoryMap
        '''
        nloc = self.mem.getLocation(va)
        if nloc is None:
            return va, 0

        nva, nvsz, nvt, nvti = nloc
        return (nva, va-nva)


class VQVivMemoryView(e_mem_qt.VQMemoryWindow, viv_base.VivEventCore):

    def __init__(self, vw, vwqgui):
        self.vw = vw
        self.vwqgui = vwqgui

        self._leading = False
        self._following = None
        self._follow_menu = None  # init'd in handler below

        e_mem_qt.VQMemoryWindow.__init__(self, vw, syms=vw, parent=vwqgui, mwname='viv')
        viv_base.VivEventCore.__init__(self, vw)

        vwqgui.addEventCore(self)
        self.mem_canvas._canv_rend_middle = True

        self.addHotKeyTarget('viv:xrefsto', self._viv_xrefsto)
        self.addHotKey('x', 'viv:xrefsto')

    def getRendToolsMenu(self):
        menu = e_mem_qt.VQMemoryWindow.getRendToolsMenu(self)
        if self.vw.server:

            leadact = QAction('lead', menu, checkable=True)

            def leadToggle():
                self._leading = not self._leading
                # We can only follow if not leading... (deep huh? ;) )
                self._follow_menu.setEnabled(not self._leading)
                if self._leading:
                    self._following = None
                    locexpr = self.addr_entry.text()
                    self.vw.iAmLeader(self.uuid, self.getEnviNavName(), locexpr)
                else:
                    self.vw.killLeaderSession(self.uuid)
                self.updateMemWindowTitle()

            def clearFollow():
                self._following = None
                self.updateMemWindowTitle()

            leadact.toggled.connect(leadToggle)
            menu.addAction(leadact)
            self._follow_menu = menu.addMenu('Follow..')
            self._follow_menu.addAction('(disable)', clearFollow)

            # add in the already existing sessions...
            for uuid, (user, fname) in self.vw.getLeaderSessions().items():

                def setFollow():
                    self._following = uuid
                    self.updateMemWindowTitle()
                    self.navToLeader()

                action = self._follow_menu.addAction('%s - %s' % (user, fname), setFollow)
                action.setWhatsThis(uuid)

        return menu

    def rendToolsSetName(self, user=None):
        menu = e_mem_qt.VQMemoryWindow.rendToolsSetName(self)
        if self.vw.server and self._leading:
            if user is None:
                user = self.vw.config.user.name
            self.vw.modifyLeaderSession(self.uuid, user, self.mwname)
        
    def getExprTitle(self):
        va = -1
        title = str(self.addr_entry.text())

        try:

            va = self.vw.parseExpression(title)
            name = self.vw.getName(va)
            if name is not None:
                title = name

        except Exception:
            title = 'expr error'

        if self._leading:
            title += ' (leading)'

        if self._following is not None:
            uuid = self._following
            user, window = self.vw.getLeaderInfo(uuid)
            title += ' (following %s %s)' % (user, window)

        return title, va

    def _getRenderVaSize(self):
        '''
        Vivisect steps in and attempts to map to locations when they exist.

        since we have a location database, let's use that to make sure we get a
        real location if it exists.  otherwise, we end up in no-man's land,
        since we rely on labels, which only exist for the base of a location.
        '''
        addr, size = e_mem_qt.VQMemoryWindow._getRenderVaSize(self)
        if addr is None:
            return addr, size

        loc = self.vw.getLocation(addr, range=False)
        if loc is None:
            return addr, size

        return loc[L_VA], size

    def initMemoryCanvas(self, memobj, syms=None):
        return VQVivMemoryCanvas(memobj, syms=syms, parent=self)

    def _viv_xrefsto(self):

        if self.mem_canvas._canv_curva is not None:
            xrefs = self.vw.getXrefsTo(self.mem_canvas._canv_curva)
            if len(xrefs) == 0:
                self.vw.vprint('No xrefs found!')
                return

            title = 'Xrefs To: 0x%.8x' % self.mem_canvas._canv_curva
            view = viv_q_views.VQXrefView(self.vw, self.vwqgui, xrefs=xrefs, title=title)
            dock = self.vwqgui.vqDockWidget(view, floating=True)
            dock.resize(800, 600)

    def loadDefaultRenderers(self):

        import envi.memcanvas.renderers as e_render

        # FIXME check endianness
        self.mem_canvas.addRenderer("bytes",    e_render.ByteRend())
        self.mem_canvas.addRenderer("u_int_16", e_render.ShortRend())
        self.mem_canvas.addRenderer("u_int_32", e_render.LongRend())
        self.mem_canvas.addRenderer("u_int_64", e_render.QuadRend())

        vivrend = viv_rend.WorkspaceRenderer(self.vw)
        self.mem_canvas.addRenderer('Viv', vivrend)
        self.mem_canvas.setRenderer('Viv')

    def _updateFunction(self, fva):
        for cbva, cbsize, cbfva in self.vw.getFunctionBlocks(fva):
            self.mem_canvas.renderMemoryUpdate(cbva, cbsize)

    def _rtmAddLeaderSession(self, uuid, user, fname):
        '''
        Add Action to RendToolsMenu (Opts/Follow) for a given session
        '''
        logger.info("_rtmAddLeaderSession(%r, %r, %r)", uuid, user, fname)

        def setFollow():
            self._following = uuid
            self.updateMemWindowTitle()
            self.navToLeader()

        action = self._follow_menu.addAction('%s - %s' % (user, fname), setFollow)
        action.setWhatsThis(uuid)

    def _rtmModLeaderSession(self, uuid, user, fname):
        '''
        Add Action to RendToolsMenu (Opts/Follow) for a given session
        '''
        logger.info("_rtmModLeaderSession(%r, %r, %r)", uuid, user, fname)
        
        action = self._rtmGetActionByUUID(uuid)
        action.setText('%s - %s' % (user, fname))
        if self._following == uuid:
            self.updateMemWindowTitle()

    def _rtmGetActionByUUID(self, uuid):
        for action in self._follow_menu.actions():
            if action.whatsThis() == uuid:
                return action

        logger.info("Unknown UUID: %r" % uuid)

    def _rtmDelLeaderSession(self, uuid):
        '''
        Remove Action item from RendToolsMenu (Opts/Follow)
        '''
        action = self._rtmGetActionByUUID(uuid)
        if action:
            logger.info("Removing %r from Follow menu (%r)" % (action, uuid))
            self._follow_menu.removeAction(action)
            if self._following == uuid:
                self._following = None
                self.updateMemWindowTitle()
        else:
            logger.warning("Attempting to remove Menu Action that doesn't exist: %r", uuid)
            self._rtmRebuild()

    def _rtmRebuild(self):
        '''
        Sync current menu system with vw.leaders
        '''
        self._follow_menu.clear()

        def clearFollow():
            self._following = None
            self.updateWindowTitle()

        self._follow_menu.addAction('(disable)', clearFollow)

        # add in the already existing sessions...
        for uuid, (user, fname) in self.vw.getLeaderSessions().items():
            self._rtmAddLeaderSession(uuid, user, fname)

    def _rtmGetUUIDs(self):
        '''
        Returns a list of active UUIDs for leader sessions
        '''
        return [action.whatsThis() for action in self._follow_menu.actions()]

    def VWE_SYMHINT(self, vw, event, einfo):
        va, idx, hint = einfo
        self.mem_canvas.renderMemoryUpdate(va, 1)

    def VWE_ADDLOCATION(self, vw, event, einfo):
        va, size, ltype, tinfo = einfo
        self.mem_canvas.renderMemoryUpdate(va, size)

    def VWE_DELLOCATION(self, vw, event, einfo):
        va, size, ltype, tinfo = einfo
        self.mem_canvas.renderMemoryUpdate(va, size)

    def VWE_ADDFUNCTION(self, vw, event, einfo):
        va, meta = einfo
        self.mem_canvas.renderMemoryUpdate(va, 1)

    def VWE_SETFUNCMETA(self, vw, event, einfo):
        fva, key, val = einfo
        self._updateFunction(fva)

    def VWE_SETFUNCARGS(self, vw, event, einfo):
        fva, fargs = einfo
        self._updateFunction(fva)

    def VWE_COMMENT(self, vw, event, einfo):
        va, cmnt = einfo
        self.mem_canvas.renderMemoryUpdate(va, 1)

    @idlethread
    def VWE_SETNAME(self, vw, event, einfo):
        va, name = einfo
        self.mem_canvas.renderMemoryUpdate(va, 1)
        for fromva, tova, rtype, rflag in self.vw.getXrefsTo(va):
            self.mem_canvas.renderMemoryUpdate(fromva, 1)

    @idlethread
    def VTE_IAMLEADER(self, vw, event, einfo):
        uuid, user, fname, locexpr = einfo
        self._rtmAddLeaderSession(uuid, user, fname)

    @idlethread
    def VTE_KILLLEADER(self, vw, event, einfo):
        uuid = einfo
        self._rtmDelLeaderSession(uuid)

    @idlethread
    def VTE_MODLEADER(self, vw, event, einfo):
        uuid, user, fname = einfo
        self._rtmModLeaderSession(uuid, user, fname)

    @idlethread
    def VTE_FOLLOWME(self, vw, event, einfo):
        uuid, expr = einfo
        if self._following != uuid:
            return
        self.enviNavGoto(expr)

    @idlethread
    def enviNavGoto(self, expr, sizeexpr='256', rend=''):
        if self._leading:
            self.vw.followTheLeader(str(self.uuid), str(expr))
        return e_mem_qt.VQMemoryWindow.enviNavGoto(self, expr, sizeexpr=sizeexpr, rend=rend)

    @idlethread
    def navToLeader(self):
        uuid = self._following
        logger.debug("navToLeader(%r)" % uuid)
        if uuid:
            expr = self.vw.getLeaderLoc(uuid)
            if not expr:
                return

            self.enviNavGoto(expr)



"""
Vtrace notitifers base classes and examples

Vtrace supports the idea of callback notifiers which
get called whenever particular events occur in the target
process.  Notifiers may be registered to recieve a callback
on any of the vtrace.NOTIFY_FOO events from vtrace.  One notifier
*may* be registered with more than one trace, as the "notify"
method is passed a reference to the trace for which an event
has occured...

"""
# Copyright (C) 2007 Invisigoth - See LICENSE file for details
import logging
import traceback

import vtrace


logger = logging.getLogger(__name__)


class Notifier(object):
    """
    The top level example notifier...  Anything which registers
    itself for trace events or tracegroup events should implement
    the notify method as shown here.
    """

    def __init__(self):
        """
        All extenders *must* call this.  Mostly because all the
        goop necessary for the remote debugging stuff...
        (if notifier is instantiated on server, all is well, if it's
        on the client it needs a proxy...)
        """
        pass

    def handleEvent(self, event, trace):
        """
        An "internal" handler so if we need to do something
        from an API perspective before calling the notify method
        we can have a good "all at once" hook
        """
        self.notify(event, trace)

    def notify(self, event, trace):
        logger.info("Got event: %d from pid %d", event, trace.getPid())


class VerboseNotifier(Notifier):
    def notify(self, event, trace):
        logger.info("PID %d - ThreadID (%d) got", trace.getPid(), trace.getMeta("ThreadId"))
        if event == vtrace.NOTIFY_ALL:
            ("WTF, how did we get a vtrace.NOTIFY_ALL event?!?!")
        elif event == vtrace.NOTIFY_SIGNAL:
            signo = trace.getCurrentSignal()
            ("vtrace.NOTIFY_SIGNAL %d (0x%08x)" % (signo, signo))
            if trace.getMeta("Platform") == "windows":
                logger.info(repr(trace.getMeta("Win32Event")))
        elif event == vtrace.NOTIFY_BREAK:
            logger.info("vtrace.NOTIFY_BREAK")
            logger.info("\tIP: 0x%08x", trace.getProgramCounter())
        elif event == vtrace.NOTIFY_SYSCALL:
            logger.info("vtrace.NOTIFY_SYSCALL")
        elif event == vtrace.NOTIFY_CONTINUE:
            logger.info("vtrace.NOTIFY_CONTINUE")
        elif event == vtrace.NOTIFY_EXIT:
            logger.info("vtrace.NOTIFY_EXIT")
            logger.info("\tExitCode: %d", trace.getMeta("ExitCode"))
        elif event == vtrace.NOTIFY_ATTACH:
            logger.info("vtrace.NOTIFY_ATTACH")
        elif event == vtrace.NOTIFY_DETACH:
            logger.info("vtrace.NOTIFY_DETACH")
        elif event == vtrace.NOTIFY_LOAD_LIBRARY:
            logger.info("vtrace.NOTIFY_LOAD_LIBRARY")
            logger.info("\tLoaded library %s", trace.getMeta('LatestLibrary'))
        elif event == vtrace.NOTIFY_UNLOAD_LIBRARY:
            logger.info("vtrace.NOTIFY_UNLOAD_LIBRARY")
        elif event == vtrace.NOTIFY_CREATE_THREAD:
            logger.info("vtrace.NOTIFY_CREATE_THREAD")
            logger.info("\tNew thread - ThreadID: %d", trace.getMeta("ThreadId"))
        elif event == vtrace.NOTIFY_EXIT_THREAD:
            logger.info("vtrace.NOTIFY_EXIT_THREAD")
            logger.info("Thread exited - ThreadID: %d", trace.getMeta("ExitThread", -1))
        elif event == vtrace.NOTIFY_STEP:
            logger.info("vtrace.NOTIFY_STEP")
        else:
            logger.warning("Unhandled vtrace event type of: %d", event)


class DistributedNotifier(Notifier):
    """
    A notifier which will distributed notifications out to
    locally registered notifiers so that remote tracer's notifier
    callbacks only require once across the wire.
    """
    # NOTE: once you turn on vtrace.NOTIFY_ALL it can't be turned back off yet.
    def __init__(self):
        Notifier.__init__(self)
        self.shared = False
        self.events = []
        self.notifiers = {}
        for i in range(vtrace.NOTIFY_MAX):
            self.notifiers[i] = []

    def notify(self, event, trace):
        self.fireNotifiers(event, trace)

    def fireNotifiers(self, event, trace):
        """
        Fire all our registerd local-notifiers
        """
        nlist = self.notifiers.get(vtrace.NOTIFY_ALL, [])
        for notifier in nlist:
            try:
                notifier.handleEvent(event, trace)
            except Exception:
                logger.error("Exception in notifier:\n%s", traceback.format_exc())

        nlist = self.notifiers.get(event, [])
        for notifier in nlist:
            try:
                notifier.handleEvent(event, trace)
            except Exception:
                logger.error("Exception in notifier:\n%s", traceback.format_exc())

    def registerNotifier(self, event, notif):
        """
        Register a sub-notifier to get the remote callback's via
        our local delivery.
        """
        nlist = self.notifiers.get(event)
        nlist.append(notif)

    def deregisterNotifier(self, event, notif):
        nlist = self.notifiers.get(event)
        nlist.remove(notif)

class LibraryNotifier(Notifier):
    def notify(self, event, trace):
        logger.info("LibraryNotifier.notify(%r, %r)", event, trace)

        # update unresolved breakpoints:
        trace._updateBreakAddresses()

        # check meta
        if hasattr(trace, 'db'):
            cfgBreakLibLoad = trace.db.config.vdb.BreakOnLibraryLoad
            cfgBreakLibInit = trace.db.config.vdb.BreakOnLibraryInit
        else:
            cfgBreakLibLoad = False
            cfgBreakLibInit = False

        #import envi.interactive as ei; ei.dbg_interact(locals(), globals())
        breakLibLoad = trace.getMeta('BreakOnLibraryLoad')
        if breakLibLoad or cfgBreakLibLoad:
            # stop this instant!
            trace.sendBreak()

        breakLibInit = trace.getMeta('BreakOnLibraryInit')
        if breakLibInit or cfgBreakLibInit:
            # add Breakpoint for __entry
            libnormname = trace.getMeta('LatestLibraryNorm')
            entryname = "%s.__entry" % (libnormname)
            logger.debug("BreakOnLibraryInit: %r\t\thooking %s", libnormname, entryname)

            # WARNING: this expects all libraries (and binaries) to have a 
            # __entry.  every library *does*, we just need to make sure Viv/
            # Vtrace names them appropriately.
            try:
                initva = trace.parseExpression(entryname)
                logger.warning("LoadLibrary(%r): Breakpoint added at 0x%x (%r)", libnormname, initva, entryname)
                self._doAddBreakByExp(trace, entryname)

            except Exception as e:
                logger.warning("LoadLibrary(%r): Can't add breakpoint!  %r", libnormname, e)

    def _doAddBreakByExp(self, trace, expr):
        logger.debug("_doAddBreakByExp(%r, %r)", trace, expr)
        trace.addBreakByExpr(expr)



"""
All the exception types raised by workspace APIs go here.
"""


class BlobArchException(Exception):
    def __init__(self):
        Exception.__init__(self, 'Blob loader *requires* arch option (-O viv.parsers.blob.arch="<archname>")')


class BadOutInstruction(Exception):
    def __init__(self, va):
        Exception.__init__(self, 'Hit out instruction at 0x%.8x' % va)


class InvalidLocation(Exception):
    def __init__(self, va, msg=None):
        Exception.__init__(self, 'Invalid Location 0x%.8x: %s' % (va, msg))


class DuplicateName(Exception):
    def __init__(self, origva, newva, name):
        Exception.__init__(self, 'Duplicate Name: %s at 0x%.8x and 0x%.8x' % (name, origva, newva))


class InvalidVaSet(Exception):
    def __init__(self, name):
        Exception.__init__(self, 'Invalid Va Set Specified: %s' % name)


class InvalidFunction(Exception):
    def __init__(self, va):
        Exception.__init__(self, 'VA 0x%.8x is not a function' % va)


class InvalidCodeBlock(Exception):
    def __init__(self, va):
        Exception.__init__(self, 'VA 0x%.8x is not in a code block!' % va)


class BadOpBytes(Exception):
    def __init__(self, va):
        Exception.__init__(self, 'Hit known badop bytes at va 0x%.8x ' % va)


class UnknownCallingConvention(Exception):
    def __init__(self, fva, cc=None):
        Exception.__init__(self, 'Function 0x%.8x has unknown CallingConvention: %s' % (fva, cc))


class InvalidWorkspace(Exception):
    """
    Raised when a storage module is given bunk data for loading
    a workspace.
    """
    def __init__(self, nameinfo, errinfo):
        Exception.__init__(self, 'Failed to load %s: %s' % (nameinfo, errinfo))


class ArchModDefException(Exception):
    def __init__(self, arch):
        Exception.__init__(self, "Architecture Module not defined for %s yet!" % arch)
        self.arch = arch

class InvalidArchitecture(Exception):
    def __init__(self, fileformat, arch):
        Exception.__init__(self, 'Architecture %s is not supported for %s' % (arch, fileformat))
        self.fileformat = fileformat
        self.arch = arch


class CorruptFile(Exception):
    def __init__(self, fileformat, message):
        super(CorruptFile, self).__init__('%s: corrupt file: %s' % (fileformat, message))
        self.fileformat = fileformat
        self.message = message


class CorruptPeFile(CorruptFile):
    def __init__(self, message):
        super(CorruptPeFile, self).__init__("PE", message)


class SymIdxNotFoundException(Exception):
    def __repr__(self):
        return "getSymIdx cannot determine the Index register"

class NoComplexSymIdxException(Exception):
    def __init__(self, sc=None):
        self.sc = sc
        Exception.__init__(self)

    def __repr__(self):
        return "getComplexIdx cannot determine the Index register"





'''
Calling convention and API definitions for various APIs/archs.
'''
import sys


class ImportApi:

    def __init__(self):
        self._api_lookup = {}
        self._apitype_lookup = {}

    def getImpApiType(self, tname):
        return self._apitype_lookup.get(tname)

    def updateApiDef(self, apidict):
        self._api_lookup.update(apidict)

    def getImpApi(self, funcname):
        '''
        An API definition consists of the following:
            ( rettype, retname, callconv, funcname, ( (argtype, argname), ...) )
        '''
        return self._api_lookup.get(funcname.lower())

    def getImpApiCallConv(self, funcname):
        ret = self._api_lookup.get(funcname.lower())
        if ret is None:
            return None
        return ret[2]

    def getImpApiArgs(self, funcname):
        ret = self._api_lookup.get(funcname.lower())
        if ret is None:
            return None
        return ret[4]

    def getImpApiRetType(self, funcname):
        ret = self._api_lookup.get(funcname.lower())
        if ret is None:
            return None
        return ret[0]

    def getImpApiRetName(self, funcname):
        ret = self._api_lookup.get(funcname.lower())
        if ret is None:
            return None
        return ret[1]

    def getImpApiArgTypes(self, funcname):
        ret = self._api_lookup.get(funcname.lower())
        if ret is None:
            return None
        return [argt for (argt, argn) in ret[4]]

    def getImpApiArgNames(self, funcname):
        ret = self._api_lookup.get(funcname.lower())
        if ret is None:
            return None
        return [argn for (argt, argn) in ret[4]]

    def addImpApi(self, api, arch):
        api = api.lower()
        arch = arch.lower()
        modname = 'vivisect.impapi.%s.%s' % (api, arch)
        __import__(modname)
        mod = sys.modules[modname]
        self._api_lookup.update(mod.api)
        self._apitype_lookup.update(mod.apitypes)


def getImportApi(api, arch):
    impapi = ImportApi()
    impapi.addImpApi(api, arch)
    return impapi



"""
generic workspace analysis module to seek through the undiscovered
country looking for pointers to interesting things.

in a previous life, this analysis code lived inside VivWorkspace.analyze()
"""
import logging

import envi.exc as e_exc
import visgraph.exc as g_exc
import visgraph.graphcore as g_core
from vivisect.const import RTYPE_BASEPTR, LOC_POINTER

logger = logging.getLogger(__name__)

def analyze(vw):

    logger.info('...analyzing pointers.')

    done = {}

    # Let's analyze Relocations we know are pointers
    for rva, rtype in vw.reloc_by_va.items():
        if rtype != RTYPE_BASEPTR:
            continue

        for xfr, xto, xtype, xinfo in vw.getXrefsFrom(rva):
            logger.debug('pointer(xref): 0x%x -> 0x%x', xfr, xto)
            try:
                vw.analyzePointer(xto)
            except e_exc.SegmentationViolation:
                continue
            done[xfr] = xto

    # Now, we'll analyze the pointers placed by the file wrapper (ELF, PE, MACHO, etc...)
    for pva, tva, fname, pname in vw.getVaSetRows('PointersFromFile'):
        if vw.getLocation(pva) is None:
            if tva is None:
                logger.debug('making pointer(fromFile) 0x%x (%r)', pva, pname)
            else:
                logger.debug('making pointer(fromFile) 0x%x -> 0x%x (%r)', pva, tva, pname)
            vw.makePointer(pva, tva, follow=True)
            done[pva] = tva

    for lva, lsz, lt, li in vw.getLocations(LOC_POINTER):
        tva = vw.readMemoryPtr(lva)
        if not vw.isValidPointer(tva):
            continue

        if vw.getLocation(tva) is not None:
            # event if it's a pointer, we may not have made a name for it
            if vw.getName(lva) is None:
                done[lva] = tva
            continue

        logger.debug('following previously discovered pointer 0x%x -> 0x%x', lva, tva)
        try:
            logger.debug('pointer: 0x%x -> 0x%x', lva, tva)
            vw.followPointer(tva)
            done[lva] = tva
        except Exception as e:
            logger.error('followPointer() failed for 0x%.8x (pval: 0x%.8x) (err: %s)', lva, tva, e)

    # Now, lets find likely free-hanging pointers
    for addr, pval in vw.findPointers():
        if vw.isDeadData(pval):
            continue
        try:
            logger.debug('make pointer(found): 0x%x -> 0x%x', addr, pval)
            vw.makePointer(addr, follow=True)
            done[addr] = pval
        except Exception as e:
            logger.error('makePointer() failed for 0x%.8x (pval: 0x%.8x) (err: %s)', addr, pval, e)

    # link the possible pointers up
    pgraph = g_core.Graph()
    while done:
        ptr, tgt = done.popitem()
        try:
            pgraph.addNode(ptr)
        except g_exc.DuplicateNode:
            pass
        try:
            pgraph.addNode(tgt)
        except g_exc.DuplicateNode:
            pass
        pgraph.addEdgeByNids(ptr, tgt)

    leafs = [node for node in pgraph.getNodes() if pgraph.isLeafNode(node)]
    for leaf in leafs:
        va, props = leaf
        name = vw.getName(va)
        # if there's not a name at the bottom, not work making ptr names
        if not name:
            continue
        links = [(ptr, tgt) for edge, ptr, tgt, props in pgraph.getRefsToByNid(va)]
        while links:
            ptr, tgt = links.pop(0)
            pname = vw.getName(ptr)
            if pname:
                continue
            if not vw.isLocType(ptr, LOC_POINTER):
                continue

            tgtname = vw.getName(tgt)
            if tgtname:
                name = vw._addNamePrefix(tgtname, tgt, 'ptr', '_') + '_%.8x' % ptr
                logger.debug('0x%x: adding name prefix: %r  (%r)', tgt, tgtname, name)
                vw.makeName(ptr, name)

            for edge, n1, n2, props in pgraph.getRefsToByNid(ptr):
                links.append((n1, n2))



"""
msp430 module
"""

############
# Author: Don C. Weber
# Started: 05/23/2009

import envi

from envi.archs.msp430.regs import *
from envi.archs.msp430.disasm import *
from envi.archs.msp430.const import *

class Msp430Module(envi.ArchitectureModule):

    def __init__(self):
        envi.ArchitectureModule.__init__(self, "msp430", maxinst=4)
        self._arch_dis = Msp430Disasm()

    def archGetRegCtx(self):
        return Msp430RegisterContext()

    def archGetNopInstr(self):
        return b'\x03\x43' # NOP is emulated with: MOV #0, R3

    def initRegGroups(self):
        envi.ArchitectureModule.initRegGroups(self)
        self._regGrps.update({'general': registers})

    def getPointerSize(self):
        return 2

    def pointerString(self, va):
        return '0x{:04x}'.format(va)

    def archParseOpcode(self, bytes, offset=0, va=0, extra=None):
        return self._arch_dis.disasm(bytes, offset, va)

    def getEmulator(self):
        return Msp430Emulator()

    def getArchDefaultCall(self):
        return 'msp430call'

# NOTE: This one must be after the definition of Msp430Module
from envi.archs.msp430.emu import *



'''
Finds and connects Switch Cases from Microsoft and GCC, and theoretically others,
which use pointer arithetic to determine code path for each case.

This will not connect switch cases which are actually explicit cmp/jz in the code.
'''
import sys
import time

import logging
logger = logging.getLogger(__name__)

import envi
import envi.exc as e_exc
import envi.bits as e_bits
import envi.archs.i386 as e_i386

import vivisect
import vivisect.exc as v_exc
import vivisect.cli as viv_cli
import vivisect.tools.graphutil as viv_graph
import vivisect.symboliks.emulator as vs_emu
import vivisect.symboliks.analysis as vs_anal
import vivisect.symboliks.substitution as vs_sub
import vivisect.analysis.generic.codeblocks as vagc

from vivisect.symboliks.common import *
from vivisect.tools.graphutil import PathForceQuitException


'''
this analysis module takes a two stage approach to identifying and wiring up switch cases.

the first phase is to identify the index range for the switch-case.  many switch cases are
split into logically grouped cases, such as if the case statements are "1,2,3,4,5,6,7,10,1000,
1001,1002,1005,2000,...." the compiler might break this up into three distinct dynamic branches,
one to handle the cases 1-10, another for the 1001-1005, and another to handle the 2000's.  this
first "discovery" phase identifies how many cases are handled in a given dynamic branch, as well
as the offset.  thus, for the 1000's, there may be 5 cases, and their offset would be 1001.

the second phase actually wires up the switch case instance, providing new codeflow analysis where
necessary, new codeblocks, and xrefs from the dynamic branch to the case handling code.  at the
end, names are applied as appropriate.
'''

# TODO: make all switchcase analysis for a given function cohesive (ie. don't finish up with naming until the entire function has been analyzed).  this will break cohesion if we use the CLI to add switchcases, but so be it.  (collisions in named targets... shared between jmp's in same switch/function)
# TODO: overlapping case names...  doublecheck the naming algorithm.  ahh... different switch-cases colliding on handlers?  is this legit?  or is it because we don't stop correctly?
# CHECK: are the algorithms for "stopping" correct?  currently getting 1 switch case for several cases in libc-32.  (enhance "don't analyze" checks (like, already analyzed?)  or no?)
# TODO: regrange description of Symbolik Variables... normalize so "eax" and "eax+4" make sense.


class TrackingSymbolikEmulator(vs_anal.SymbolikFunctionEmulator):
    '''
    TrackingSymbolikEmulator has two modifications from a standard SymbolikEmulator:
    * It tracks reads.  where they occur, where they read from, and
    returns as much info as possible if not discrete data.

    * If a read is from a real memory map, return that memory.

    These are both very useful for analyzing switchcases, to determine where in
    memory a read occurs and the data comes from.
    '''
    def __init__(self, vw):
        vs_anal.SymbolikFunctionEmulator.__init__(self, vw)
        self.__width__ = vw.psize
        self.clear()

    def clear(self):
        '''
        Clear read-tracker to start another analysis
        '''
        self._trackReads = []

    def track(self, va, addrval, val):
        '''
        Track reads in this SymbolikEmulator for later analysis
        '''
        self._trackReads.append((va, addrval, val))

    def getTrackInfo(self):
        '''
        Retrieve reads so far, for analysis
        '''
        return self._trackReads

    def setupFunctionCall(self, fva, args=None):
        logger.info("setupFunctionCall: SKIPPING!")

    def readSymMemory(self, symaddr, symsize, vals=None):
        '''
        The readSymMemory API is designed to read from the given
        symbolik address for the specified symbolik length.  If the
        current symbolik emulator has no knowledge of the state of
        the given memory symbol, None is returned.
        '''
        addrval = symaddr.update(self).solve(emu=self, vals=vals)
        logger.debug('   TSE.readSymMemory: symaddr:%r  addrval:0x%x symsize:%s',symaddr, addrval, symsize)

        # check for a previous write first...
        symmem = self._sym_mem.get(addrval)
        if symmem is not None:
            symaddr, symval = symmem
            self.track(self.getMeta('va'), symaddr, symval)
            return symval

        # If we have a workspace, check it for meaningful symbols etc...
        if self._sym_vw is not None:
            # Make a special check for imports...
            loc = self._sym_vw.getLocation(addrval)
            if loc is not None:
                lva, lsize, ltype, linfo = loc
                if ltype == vivisect.LOC_IMPORT:
                    # return name of import
                    symval = Var(linfo, self.__width__)
                    self.track(self.getMeta('va'), symaddr, symval)
                    return symval

            if self._sym_vw.isValidPointer(addrval):
                if symsize.isDiscrete():
                    size = symsize.solve()
                else:
                    size = symsize

                if size in (1, 2, 4, 8):
                    # return real number from memory
                    fmt = e_bits.getFormat(size, self._sym_vw.getEndian(), signed=True)
                    val, = self._sym_vw.readMemoryFormat(addrval, fmt)
                    self.track(self.getMeta('va'), symaddr, val)
                    return Const(val, size)

                # return string  (really?)
                symval = self._sym_vw.readMemory(addrval, size)
                self.track(self.getMeta('va'), symaddr, symval)
                return symval

        return Mem(symaddr, symsize)


def contains(symobj, subobj):
    '''
    search through an AST looking for a particular type of thing.

    args:
        symobj - the AST to walk through
        subobj - the value to compare against (symobj.solve())

    returns:
        contains - whether or not the AST contains the sought value
        path - the operator path to get to that object (last one seen)
    '''
    def _cb_contains(path, symobj, ctx):
        '''
        walkTree callback for determining presence within an AST
        '''
        symsolve = symobj.solve()
        # logger.debug("==--== %r   %r   %r", repr(symobj), symsolve, ctx['compare'])
        if symsolve == ctx['compare']:
            ctx['contains'] = True
            ctx['path'] = tuple(path)

        return symobj

    ctx = {'compare':   subobj.solve(),
           'contains':  False,
           'path':      ()
           }

    symobj.walkTree(_cb_contains, ctx)
    return ctx['contains'], ctx['path']

def getUnknowns(symvar):
    '''
    determine unknown registers in this symbolik object
    '''
    def _cb_grab_vars_n_mems(path, symobj, ctx):
        '''
        walkTree callback for grabbing Var objects
        '''
        #logging.debug("... unk: %r", symobj)
        if symobj.symtype == SYMT_VAR:
            varlist = ctx.get(SYMT_VAR)
            if symobj.name not in varlist:
                varlist.append(symobj.name)

        # sometimes the index comes from a global or local variable...
        if symobj.symtype == SYMT_MEM:
            memlist = ctx.get(SYMT_MEM)
            if symobj not in memlist:
                memlist.append(symobj)

    unks = {SYMT_VAR:[], SYMT_MEM:[]}
    symvar.walkTree(_cb_grab_vars_n_mems, unks)
    return unks

def replaceObj(ast, symobj, symreplacement):
    '''
    Walk the AST looking for symobj, replace with symreplacement
    '''
    def _cb_check_and_replace(path, symobj, ctx):
        searchobj = ctx.get('searchobj')
        for kidx, kid in enumerate(symobj.kids):
            if searchobj == kid:
                replacement = ctx.get('replacewith')
                symobj.setSymKid(kidx, replacement)
                ctx['success'] = True

    ctx = {
        'searchobj': symobj,
        'replacewith': symreplacement,
        'success': False,
    }

    ast.walkTree(_cb_check_and_replace, ctx)
    return ctx.get('success')

def hasMul(symobj):
    '''
    Returns True or False based on whether the symobj AST has a "Mul" object
    '''
    def _cb_check_for_mul(path, symobj, ctx):
        if symobj.symtype == SYMT_OPER_MUL:
            ctx['inthere'] = True

    ctx = {'inthere': False}
    symobj.walkTree(_cb_check_for_mul, ctx)
    return ctx['inthere']

def peelIdxOffset(symobj):
    '''
    Peel back ignorable wrapped layers of a symbolik Index register, and track
    offset in the process.  Once we've skipped out of the ignorable symobj's we
    should end up with something normalized-ish.
    '''
    offset = 0
    while True:
        # peel off o_subs and size-limiting o_ands and o_sextends
        if isinstance(symobj, o_and) and symobj.kids[1].isDiscrete() and symobj.kids[1].solve() in e_bits.u_maxes:
            # this wrapper is a size-limiting bitmask
            pass

        elif isinstance(symobj, o_sub):
            # this is an offset, used to rebase the index into a different pointer array
            if symobj.kids[1].isDiscrete():
                offset += symobj.kids[1].solve()

        elif isinstance(symobj, o_add):
            # this is an offset, used to rebase the index into a different pointer array
            if symobj.kids[1].isDiscrete():
                offset -= symobj.kids[1].solve()

        elif isinstance(symobj, o_sextend):
            # sign-extension is irrelevant for indexes
            pass

        # anything else and we're done peeling
        else:
            break

        # this algorithm depends on only taking left turns
        symobj = symobj.kids[0]

    return symobj, offset

# TODO: possibly only take calls in the first codeblock?
def targetNewFunctions(vw, fva):
    '''
    scan through all direct calls in this function and force analysis of called functions
    if this is too cumbersome, we'll just do the first one, or any in the first codeblock
    '''
    done = []
    todo = list(vw.getFunctionBlocks(fva))

    while len(todo):
        cbva, cbsz, cbfva = todo.pop()
        endva = cbva + cbsz
        while cbva < endva:
            try:
                op = vw.parseOpcode(cbva)

            except:
                logger.warning('failed to parse opcode within function at 0x%x (fva: 0x%x)', cbva, fva)
                cbva += 1
                continue

            for xrfr, xrto, xrt, xrflag in vw.getXrefsFrom(cbva):
                if not (op.iflags & envi.IF_CALL):
                    continue

                # TODO: improve upon determination of call
                tgtva = op.getOperValue(0)
                #logger.debug("-- 0x%x", tgtva)
                if tgtva is None or not vw.isValidPointer(tgtva):
                    continue

                if vw.getFunction(tgtva) is not None:
                    continue

                if tgtva in done:
                    continue
                done.append(tgtva)

                logger.info('Making new function: 0x%x (from 0x%x)', tgtva, xrfr)
                vw.makeFunction(tgtva)

            cbva += len(op)


THUNK_BX_CALL_LEN = 5


class ThunkReg:
    def __init__(self, regname, tgtval):
        self.regname = regname
        self.tgtval = tgtval

    def thunk_reg(self, emu, fname, symargs):
        '''
        These thunk_reg functions return the "RETURN" address, which the calling
        function then adds the appropriate amount to obtain the .got.plt base

        This simply applies the RET address to the appropriate register for 
        symbolik analysis
        '''
        vw = emu._sym_vw
        rctx = vw.arch.archGetRegCtx()
        regval = emu.getMeta('calling_va')
        oploc = vw.getLocation(regval)
        if oploc is None:
            regval += THUNK_BX_CALL_LEN 
        else:
            regval += oploc[L_SIZE]

        regobj = Const(regval, vw.psize)
        reg = rctx.getRealRegisterName(self.regname)
        logger.debug("thunk_%s is being called! %s\t%s\t%s", reg, emu, symargs, regobj)
        emu.setSymVariable(reg, regobj)


UNINIT_CASE_INDEX = -2


class SwitchCase:
    '''
    Actually do the analysis for each dynamic branch.
    * First, for analysis: target addresses, case-offset, number of cases
    * Second, to wire things up.

    Tracks the state of analysis so each component can easily do it's part
    without handing a ton of variables around between functions.
    '''
    def __init__(self, vw, jmpva, timeout=180):
        self.vw = vw
        self.jmpva = jmpva
        self.timeout = timeout
        self.jmpop = vw.parseOpcode(jmpva)
        logger.info('=== analyzing SwitchCase branch at 0x%x: %r ===' % (jmpva, self.jmpop))

        self.sctx = vs_anal.getSymbolikAnalysisContext(vw)
        if self.sctx is None:
            raise e_exc.ArchNotImplemented("Current architecture has no SymbolikTranslator")

        self.xlate = self.sctx.getTranslator()

        # 32-bit i386 thunk_reg handling.  this should be the only oddity for this architecture
        for tva, regname, tgtval in vw.getVaSetRows('thunk_reg'):
            fname = vw.getName(tva, True)
            trobj = ThunkReg(regname, tgtval)
            self.sctx.addSymFuncCallback(fname, trobj.thunk_reg)
            logger.debug( "sctx.addSymFuncCallback(%s, thunk_reg)" % fname)


        self._sgraph = None
        self._codepath = None
        self._codepathgen = None


        try:
            self.max_instr_count = vw.config.viv.analysis.symswitchcase.max_instr_count
            self.max_cases = vw.config.viv.analysis.symswitchcase.max_cases
            self.case_failure = vw.config.viv.analysis.symswitchcase.case_failure
            self.min_func_instr_size = vw.config.viv.analysis.symswitchcase.min_func_instr_size

        except AttributeError as e:
            logger.warning('configuration failure, using defaults: %r', e)
            self.max_instr_count  = 10
            self.max_cases   = 500
            self.case_failure = 5000
            self.min_func_instr_size = 10

        self.clearCache()



    def clearCache(self):
        '''
        While analyzing a switchcase, we end up caching several calculated items.
        This function clears them, so we can reset the state, like to skip
        inappropriate paths (eg.   if x> 3 and if x == 0
        '''
        self.cspath = None
        self.aspath = None
        self.fullpath = None

        self.upper = None
        self.lower = None
        self.count = None
        self.baseoff = None
        self.baseIdx = None

        self.jmpsymvar = None


    def analyze(self):
        '''
        Generic "analyze" function, which simply calls makeSwitch()
        '''
        self.makeSwitch()

    #### primitives for switchcase analysis ####
    def getJmpSymVar(self):
        '''
        returns the Simple Symbolik state of the dynamic target of the branch/jmp
        this is the symbolik register at the start of the last codeblock.
        '''
        if self.jmpsymvar is not None:
            return self.jmpsymvar

        self.jmpsymvar = self.xlate.getOperObj(self.jmpop, 0)
        return self.jmpsymvar

    def getSymTarget(self):
        '''
        Returns the Symbolik state of the dynamic target of the branch/jmp
        short indicates just in the context of the last codeblock.
        not short indicates the context at the start of the last codeblock (from funcva)
        '''
        jmpsymvar = self.getJmpSymVar()

        cspath, aspath, fullpath = self.getSymbolikParts()
        emu = aspath[0]
        tgtsym = jmpsymvar.update(emu)
        return tgtsym


    def getSymIdx(self):
        '''
        returns the symbolik index register, and the type of object
        returns first one found, first checking SYMT_VAR's and only
        then looking at SYMT_MEM symbolic indexes
        '''
        symtgt = self.getSymTarget()
        unks = getUnknowns(symtgt)

        cspath, aspath, fullpath = self.getSymbolikParts()

        for unk in unks.get(SYMT_VAR):
            unkvar = cspath[0].getSymVariable(unk)
            if unkvar.isDiscrete():
                continue
            return (SYMT_VAR, unk)

        for unk in unks.get(SYMT_MEM):
            unkvar = cspath[0].getSymVariable(unk)
            if unkvar.isDiscrete():
                continue
            return (SYMT_MEM, unk)

        raise v_exc.SymIdxNotFoundException(cspath, aspath)

    def getConstraints(self):
        '''
        Returns a list of Constraint objects from the current codepath.
        Some of these constraints should have checks for boundaries against our
        index variable which will prove helpful determining case-offset and
        case-count.
        '''
        csp, asp, fullp = self.getSymbolikParts()

        cons = [eff for eff in fullp[1] if eff.efftype==EFFTYPE_CONSTRAIN]
        [con.reduce() for con in cons]
        return cons

    def getBoundingCons(self, cplxIdx):
        return [con for con in self.getConstraints() if contains(con, cplxIdx)[0] ]

    def getSymbolikJmpBlock(self, emuclass=vs_anal.SymbolikFunctionEmulator):
        '''
        Returns symbolik effects for the last codeblock, which should never change.
        This should be small and does not require a path be generated, so it should
        be safe for early analysis and vetting of the dynamic branch (if it's not
        a switchcase, don't do all the heavy lifting)
        '''
        vw = self.vw
        sctx = self.sctx
        jmpva = self.jmpva

        fva = vw.getFunction(jmpva)
        cb = vw.getCodeBlock(jmpva)
        if cb is None:
            raise v_exc.InvalidCodeBlock("Dynamic Branch is not currently part of a CodeBlock!")
        cbva, cbsz, cbfva = cb

        if self._sgraph is None:
            self._sgraph = sctx.getSymbolikGraph(fva)

        jnva, jnode = self._sgraph.getNode(cbva)

        symemu = emuclass(vw)
        aeffs = symemu.applyEffects(jnode.get('symbolik_effects'))

        #TODO: ? cache these?  depends on how widely we use this function.
        return jnode, symemu, aeffs

    def getSymbolikParts(self, doNext=False):
        '''
        Puts together a path from the start of the function to the jmpva, breaking off the last 
        codeblock, and returning the symbolik path for both sections:
        cspath = (emu, effs) for the "context" path (from function start)
        aspath = (emu, effs) for the last codeblock

        if "next" isn't False, the next iteration is returned.  otherwise, the current iteration is (same as last call).
        this allows the Symbolik Parts to be easily accessible from various parts of the algorithm without handing around
        a lot of state as function args.
        '''
        if not doNext and self.cspath is not None and self.aspath is not None and self.fullpath is not None:
            return self.cspath, self.aspath, self.fullpath

        self.clearCache()

        vw = self.vw
        sctx = self.sctx
        jmpva = self.jmpva

        fva = vw.getFunction(jmpva)
        cb = vw.getCodeBlock(jmpva)
        if cb is None:
            raise Exception("Dynamic Branch is not currently part of a CodeBlock!")
        cbva, cbsz, cbfva = cb

        if self._sgraph is None:
            self._sgraph = sctx.getSymbolikGraph(fva)

        if self._codepathgen is None:
            pathGenFactory = viv_graph.PathGenerator(self._sgraph)
            self._codepathgen = pathGenFactory.getFuncCbRoutedPaths(fva, cbva, 1, timeout=self.timeout)

        self._codepath = next(self._codepathgen)
        contextpath = self._codepath[:-1]
        analpath = self._codepath[-1:]

        self.cspath = next(sctx.getSymbolikPaths(fva, graph=self._sgraph, args=None, paths=[contextpath]))
        self.aspath = next(sctx.getSymbolikPaths(fva, graph=self._sgraph, args=[], paths=[analpath]))

        if len(self.cspath[1]) == 0:
            self.cspath = self.aspath

        self.fullpath = next(sctx.getSymbolikPaths(fva, graph=self._sgraph, args=None, paths=[self._codepath]))

        return self.cspath, self.aspath, self.fullpath

    def getComplexIdx(self):
        '''
        Complex Index is the Applied Symboliks state (of the context-path) for
        the simple index returned by getSymIdx()

        eg.  if the simple symbolik index is `Var('edx', 4)` but the
            complex could be `Arg(2, 4)` or something similar
        '''
        symtype, smplIdx = self.getSymIdx()

        (csemu, cseffs), asp, fullp = self.getSymbolikParts()
        if symtype == SYMT_VAR:
            cplxIdx = csemu.getSymVariable(smplIdx)

        elif symtype == SYMT_MEM:
            symloc, symsz = smplIdx.kids
            cplxIdx = csemu.readSymMemory(symloc.update(csemu), symsz)

        return cplxIdx

    #### mid-level functionality ####
    def getBaseSymIdx(self):
        '''
        (cached)
        returns the baseIdx and greatest index offset (ie.  the offset at the point of jmp).
        BETTER EXPLAIN WHAT BaseSymIdx is.
        '''
        if None not in (self.baseIdx, self.baseoff):
            return self.baseIdx, self.baseoff

        def _cb_peel_idx(path, symobj, ctx):
            '''
            This is for troubleshooting and analysis only.
            '''
            logger.debug( ' PATH: %r\n SYMOBJ: %r\n CTX: %r\n' % (path, symobj, ctx))

        def _cb_mark_longpath(path, symobj, ctx):
            '''
            Determine a long path through an AST.
            Long paths mean we have the most complete subcomponent
            '''
            #logger.debug( ' PATH: %r\n SYMOBJ: %r\n CTX: %r\n' % (path, symobj, ctx))
            longpath = ctx.get('longpath')
            if longpath is None:
                ctx['longpath'] = list(path)
            elif len(path) > len(longpath):
                ctx['longpath'] = list(path)

        (csemu, cseff), aspath, fullpath = self.getSymbolikParts()

        idx = self.getComplexIdx()
        if idx is None:
            raise v_exc.NoComplexSymIdxException(self)

        idx.reduce()

        # peel it back
        ctx = {}
        idx.walkTree(_cb_mark_longpath, ctx)
        longpath = ctx.get('longpath')

        # now compare the long path against constraints that contain it.  find sweet spot
        count = last = None
        which = None
        offset = 0

        # peel back the constraints to remove any incidental modifications to the index variable
        #  some compilers separate disparate groupings of cases into multiple dynamic branches.
        #  in order to make all pointer/offset arrays 0-based, the index gets modified (subtracted)


        # HACK: this is left-handed, and based on the longpath.  it may also benefit from actual comparing with the 
        # known index.
        #logging.info("LONGPATH: " + '\n'.join([repr(x) for x in longpath]))
        for symobj in longpath:
            last = count
            count = len(self.getBoundingCons(symobj))
            logger.debug(" longpath constraints %d: %r" % (count, symobj))

            # peel off o_subs and size-limiting o_ands and o_sextends
            if isinstance(symobj, o_and) and symobj.kids[1].isDiscrete() and symobj.kids[1].solve() in e_bits.u_maxes:
                mask = symobj.kids[1].solve()
                pass    # this wrapper is a size-limiting bitmask

            elif isinstance(symobj, o_sub):
                offset += symobj.kids[1].solve() # this is an offset, used to rebase the index into a different pointer array

            elif isinstance(symobj, o_add):
                offset -= symobj.kids[1].solve() # this is an offset, used to rebase the index into a different pointer array

            elif isinstance(symobj, o_sextend):
                pass # sign-extension is irrelevant for indexes

            # anything else and we're done peeling
            else:
                break

        logger.debug(" longpath DONE: (%r) %r" % (offset, symobj))
        # now figure what was cut, and what impact it has.
        self.baseIdx = symobj
        self.baseoff = offset

        return symobj, offset

    def getNormalizedConstraints(self):
        '''
        takes in a complex index symbolik state
        returns a list of tuples of (SYMT_CON_??, offset)
        '''
        retcons = []
        baseIdx, baseoff = self.getBaseSymIdx()
        SKIPS = (SYMT_CON_GT, SYMT_CON_GE, SYMT_CON_LT, SYMT_CON_LE)

        for con in self.getBoundingCons(baseIdx): # merge the two
            cons = con.cons

            if not cons.symtype in SKIPS:
                #logger.debug("SKIPPING NON-LIMITING: cons = %s", repr(cons))
                #return None, None, None
                continue

            if cons.kids[1].symtype == SYMT_CONST:
                symvar = cons.kids[0]
                symcmp = cons.kids[1]

            elif cons.kids[0].symtype == SYMT_CONST:
                symcmp = cons.kids[0]
                symvar = cons.kids[1]

            else:
                # neither side of the constraint is a CONST.  this constraint does 
                # not set static bounds on idx
                #logger.debug("SKIPPING (non-discrete): cons = %s", repr(cons))
                continue

            if not contains(symvar, baseIdx):
                logger.debug("Ignoring Constraint not based on Index: %r" % cons)
                continue

            if cons.symtype == SYMT_CON_LT and symcmp.solve() == 0:
                #logger.debug("SKIPPING Constraint Checking for Zero: %r" % cons)
                continue

            conthing, consoff = peelIdxOffset(symvar)

            if conthing != baseIdx:
                logger.debug("Constraint not bound-limiting for our index: %r  != %r" % (conthing, baseIdx))
                continue

            logger.debug("Valid bound-limiting Constraint: %r\n\t%r\n\t%r\t%r + %r" % (cons, symvar, conthing, consoff, symcmp))
            retcons.append((con, cons.symtype, consoff+symcmp.solve()))

        return retcons

    def getBounds(self):
        '''
        Determine the lower index, the upper index, and the offset.
        lower and upper are inclusive.
        '''
        if None not in (self.lower, self.upper, self.baseoff):
            return self.lower, self.upper, self.baseoff

        (csemu, cseffs), asp, fullp = self.getSymbolikParts()

        lower = 0       # the smallest index used.  most often wants to be 0
        upper = None    # the largest index used.  max=self.max_cases
        count = 0
        baseoff = None

        try:
            while lower is None or (lower == 0 and upper is None) or (upper is not None and upper <= lower): # note: this will fail badly when it fails.  make this dependent on the codepathgen
                # get the index we'll be looking for in constraints
                lower = 0           # the smallest index used.  most often wants to be 0
                upper = None        # the largest index used.  max=self.max_cases

                if count != 0:
                    (csemu, cseffs), asp, fullp = self.getSymbolikParts(doNext=True)

                count += 1

                ##### PEAL the cplxIdx #####
                baseIdx, baseoff = self.getBaseSymIdx()
                logger.debug("baseIdx: %r", baseIdx)

                # there are two important offsets: constraint offsets and index offsets
                #   index offsets are mostly subtractions from the actual number used (eg. index 1500 would be idx-1500 for an offset of 1500)
                #   constraint offsets are subtractions from the index at the point of the constraint check.  these are accounted for by getNormalizedConstraints()
                for con, stype, offset in self.getNormalizedConstraints():

                    #conthing, consoff = peelIdxOffset(symvar)
                    if stype == SYMT_CON_GT: # this is setting the lower bound
                        newlower = offset + 1
                        if newlower > lower:
                            logger.info("==setting a lower bound:  %s -> %s", lower, newlower)
                            lower = newlower

                    elif stype == SYMT_CON_GE: # this is setting the lower bound
                        newlower = offset
                        if newlower > lower:
                            logger.info("==setting a lower bound:  %s -> %s", lower, newlower)
                            lower = newlower

                    elif stype == SYMT_CON_LT: # this is setting the upper bound
                        newupper = offset - 1
                        if upper is None or newupper < upper and newupper > 0:
                            logger.info("==setting a upper bound:  %s -> %s", upper, newupper)
                            upper = newupper

                    elif stype == SYMT_CON_LE: # this is setting the upper bound
                        newupper = offset
                        if upper is None or newupper < upper and newupper > 0:
                            logger.info("==setting a upper bound:  %s -> %s", upper, newupper)
                            upper = newupper

                    else:
                        logger.info("Unhandled comparator:  %s\n", repr(cons))

                logger.info("Determined bounds: %r %r\n" % (lower, upper))
        except StopIteration:
            pass

        # if upper is None:  we need to exercise upper until something doesn't make sense.  
        # we also need to make sure we don't analyze non-Switches.  
        if upper is None:
            upper = self.max_cases

        if lower < baseoff:
            # assume the compiler doesn't adjust the index below 0
            lower = baseoff

        #logger.info("Lower: %r\tUpper: %r\tOffset: %r\tIndex: %r", lower, upper, baseoff, smplIdx)
        self.upper = upper
        self.lower = lower
        self.baseoff = baseoff

        return self.lower, self.upper, self.baseoff

    def getDerefs(self):
        '''
        roll through effects and look for EFF_READMEM
        '''
        derefs = []

        csp, asp, fullp = self.getSymbolikParts()
        csemu, cseffs = csp
        asemu, aseffs = asp

        # create a tracking emulator and populate with with current "csp" state
        semu = TrackingSymbolikEmulator(self.vw)
        semu.setSymSnapshot(csemu.getSymSnapshot())

        idxtype, symIdx = self.getSymIdx()

        # set our index to 0, to get the base of pointer/offset arrays
        semu.setSymVariable(symIdx, Const(0, 8))

        for eff in aseffs:
            startlen = len(semu.getTrackInfo())

            if eff.efftype == EFFTYPE_READMEM:
                if eff.va == self.jmpva:
                    continue

                symaddr = eff.symaddr.update(emu=semu)
                if symaddr.isDiscrete():
                    solution = symaddr.solve()
                    logger.info("analyzing ptr: 0x%x->0x%x" % (eff.va, solution))

                    if not self.vw.isValidPointer(solution):
                        logger.warning(("ARRRG: Non-pointer in ReadMemory???"))

                    target = semu.readSymMemory(eff.symaddr, eff.symsize)
                    size = target.getWidth()
                    derefs.append((eff.va, symaddr, solution, target.solve(), size))

            endlen = len(semu.getTrackInfo())

        return derefs



    #### higher level functionality
    def makeSwitch(self):
        '''
        DIFFERENT TYPES OF SWITCH CASES:
        * VS's multidimensional
        * VS's single-dimension
        * Posix single dimension
        * Any of them with offset (EBX for Posix PIE, VS's base reg (typically 0x10000))

        What we need:
        * Resultant location for xrefs (iterative)
        * Index symboliks (for Constraint checking, addition/subtraction)
        * potentially, base va offset?  if only to identify the others.... but maybe not.

        How do we get from jmpreg (which is a very limited scope symboliks) to full scope symboliks run (from funcva)?
        * va effect comparison?
        '''

        vw = self.vw
        start = time.time()

        # only support branching switch-cases (ie, not calls)
        if not (self.jmpop.iflags & envi.IF_BRANCH):
            logger.info("makeSwitch: exiting: not a branch opcode: 0x%x: %r", self.jmpop.va, self.jmpop)
            return

        # skip any that already have xrefs away from (already been discovered?)
        if len(vw.getXrefsFrom(self.jmpva)):
            logger.info('skipping existing switchcase: 0x%x', self.jmpva)
            return

        # if we aren't a part of a function, where are we?
        funcva = self.vw.getFunction(self.jmpva)
        if funcva is None:
            logger.error("ERROR getting function for jmpva 0x%x", self.jmpva)
            return

        # skip when the function is the first instruction (PLT?)
        if funcva == self.jmpva:
            logger.info("skipping: function va IS jmpva 0x%x (PLT?)", self.jmpva)
            return

        # skip if insufficient instructions in the function to have an interesting switchcase.
        instrcount = vw.getFunctionMeta(funcva, 'InstructionCount')
        if instrcount < self.min_func_instr_size:
            logger.info("Ignoring jmp in too small a function: %d instructions (0x%x)", instrcount, self.jmpva)
            return

        # skip if we don't have a multiply in the jmpva calculation (always going to be an offset 
        # into a pointer array, most often of size 2+)
        jnode, jemu, jaeffs = self.getSymbolikJmpBlock()
        jmptgt = self.getJmpSymVar()
        if not hasMul(jmptgt.update(jemu)):
            logger.info("skipping: JmpSymVar doesn't have multiplication! (0x%x)", self.jmpva)
            return


        # create a tracking emulator and populate with with current "csp" state
        try:
            # relying on getBounds() to bail on non-switch-cases
            lower, upper, baseoff = self.getBounds()
            if None in (lower, ):
                logger.info("something odd in count/offset calculation...(%r,%r,%r) skipping 0x%x...",
                            lower, upper, baseoff, self.jmpva)
                return

            count = upper - lower + 1
            if count > self.max_cases:
                if count > self.case_failure:
                    logger.warning("TOO many switch cases detected: %d.  FAILURE.  Skipping this dynamic branch (0x%x)", count, self.jmpva)
                    return

                logger.warning("too many switch cases during analysis: %d   limiting to %d (0x%x)", count, self.max_cases, self.jmpva)
                count = self.max_cases

            # determine deref-ops...  uses TrackingSymbolikEmulator
            # iterCases
            cases = {}
            memrefs = []
            for idx, addr in self.iterCases():
                logger.info("0x%x analyzeSwitch: idx: %s \t address: 0x%x", self.jmpva, idx, addr)

                # determining when to stop identifying switch-cases can be tough.  we assume that we have the 
                # correct number handed into this function in "count", but currently we'll stop analyzing
                # if we run into trouble.
                if not vw.isValidPointer(addr):
                    logger.info("found invalid pointer.  quitting.  (0x%x in 0x%x)", addr, self.jmpva)
                    break

                tloc = vw.getLocation(addr)
                if tloc is not None and tloc[0] != addr:
                    # pointing at something not right.  must be done.
                    logger.info("found overlapping location at 0x%x (cur: 0x%x).  quitting. (0x%x)", \
                            addr, tloc[0], self.jmpva)
                    break

                if len(cases):
                    xrefsto = vw.getXrefsTo(addr)
                    # if there is an xref to this target from within this function, we're still ok... ish?
                    if len(xrefsto):
                        good = True
                        for xrfr, xrto, xrt, xrtinfo in xrefsto:
                            xrfrfunc = vw.getFunction(xrfr)
                            if xrfrfunc == funcva:
                                continue

                            # this one is *not* in the same function
                            good = False

                        if not good:
                            logger.info("target location (0x%x) has xrefs. (0x%x)", addr, self.jmpva)
                            break

                # this is a valid thing, we have locations...  match them up
                caselist = cases.get(addr)
                if caselist is None:
                    caselist = []
                    cases[addr] = caselist
                caselist.append( idx )

                # make the connections
                vw.addXref(self.jmpva, addr, vivisect.REF_CODE)
                nloc = vw.getLocation(addr)
                if nloc is None:
                    vw.makeCode(addr)

            # makeNames (done separately because some indexes may be combined)
            makeNames(vw, self.jmpva, cases, baseoff)

            # store some metadata in a VaSet
            vw.setVaSetRow('SwitchCases', (self.jmpva, self.jmpva, count) )

            if baseoff is not None:
                lower += baseoff
                upper += baseoff

            vw.setComment(self.jmpva, "lower: 0x%x, upper: 0x%x" % (lower, upper))

            vagc.analyzeFunction(vw, funcva)

        except StopIteration as e:
            logger.info("Switchcase Analysis BOMBED OUT (Couldn't Find Valid Path) 0x%x\n%r", self.jmpva, e)

        except v_exc.SymIdxNotFoundException as e:
            logger.info("Switchcase Analysis BOMBED OUT (SymIdx) 0x%x\n%r", self.jmpva, e)

        except v_exc.NoComplexSymIdxException as e:
            logger.info("Switchcase Analysis BOMBED OUT (SymIdx=%r \t ComplexIdx=None) 0x%x\n%r", e.sc.getSymIdx(), self.jmpva, e)

        except PathForceQuitException as e:
            logger.warning("Switchcase Analysis BOMBED OUT (Path Timeout!) 0x%x\n%r", self.jmpva, e)
            vw.setVaSetRow('SwitchCases_TimedOut', (self.jmpva, self.timeout) )

        except RuntimeError as e:
            if 'StopIteration' in repr(e):
                logger.info("Switchcase Analysis BOMBED OUT (Couldn't Find Valid Path) 0x%x", self.jmpva)
            else:
                logger.warning("Switchcase Analysis BOMBED OUT 0x%x\n%r", self.jmpva, e, exc_info=1)

        except Exception as e:
            logger.warning("Switchcase Analysis BOMBED OUT 0x%x\n%r", self.jmpva, e, exc_info=1)

        logger.debug("Symswitchcase took %.3f for jump at virtual address 0x%x", time.time() - start, self.jmpva)


    def markDerefs(self):
        '''
        Apply xrefs for derefs (the tables) and mark pointers or numbers (for offset-based switches)
        '''
        lower, upper, offset = self.getBounds()

        for va, symaddr, addr, tgt, sz in self.getDerefs():
            # first make the xref from opcode to data:
            self.vw.addXref(va, addr, REF_DATA)

            if sz == self.vw.psize and self.vw.isValidPointer(tgt):
                for count in range(upper-lower+1):
                    self.vw.makePointer(addr + (sz*count))

            else:
                for count in range(upper-lower+1):
                    self.vw.makeNumber(addr + (sz*count), sz)

    def iterCases(self):
        '''
        Generator which yields the case index and the target address
        '''
        vw = self.vw

        (csemu, cseffs), aspath, fullpath = self.getSymbolikParts()

        idxtype, symidx = self.getSymIdx()
        logger.info('getSymIdx: %r/%r (0x%x)', idxtype, symidx, self.jmpva)

        jmptgt = self.getSymTarget()
        lower, upper, offset = self.getBounds()
        logger.debug( " itercases: %r %r %r %r (0x%x)", symidx, lower, upper, offset, self.jmpva)

        # use a TrackingSymbolikEmulator so we get the "readSymMemory()" function actually hitting 
        # the r/o data from our VivWorkspace
        symemu = TrackingSymbolikEmulator(vw)
        symemu.setSymSnapshot(csemu.getSymSnapshot())

        logger.debug("jmpva: 0x%x\t\tsymidx: %r", self.jmpva, symidx)   # TODO: we return either a Var name string or a SymObj (for Mem types)... should we just stick with the symbolik object?  since we're removing the ties to getSymVariable() and moving to replaceObj() and our nifty 'jmpidx' variable.
        if idxtype == SYMT_VAR:
            symidx = Var(symidx, self.vw.psize)
        replaceObj(jmptgt, symidx, Var('jmpidx', symidx.getWidth()))
        #workJmpTgt = jmptgt.update(emu=symemu)


        for idx in range(lower-offset, upper-offset+1):
            symemu.setSymVariable('jmpidx', Const(idx, 8))
            workJmpTgt = jmptgt.update(emu=symemu)  # would "jmptgt.solve(vals={'jmpidx': idx})" work?
            logger.info(" itercases: workJmpTgt: %r (0x%x)", workJmpTgt, self.jmpva)
            coderef = workJmpTgt.solve(emu=symemu, vals={symidx:idx})
            #coderef = jmptgt.update(emu=symemu)
            logger.debug(" itercases: %r, %r", idx, hex(coderef))
            yield idx + offset, coderef

def makeNames(vw, jmpva, cases, baseoff=0):
    '''
    Create names and xrefs for each identified case.
    '''
    for addr, indexes in cases.items():
        outstrings = []
        combined = False
        start = last = UNINIT_CASE_INDEX
        for case in indexes:
            if case == last+1:
                if not combined:
                    combined = True
                    start = last
            else:
                if combined:
                    combined = False
                    outstrings.append("%Xto%X" % (start+baseoff, last+baseoff))
                elif last != UNINIT_CASE_INDEX:
                    outstrings.append("%X" % (last+baseoff))
            last = case

        # catch the last one if highest cases are combined
        if combined:
            combined = False
            outstrings.append("%Xto%X" % (start+baseoff, last+baseoff))
        else:
            outstrings.append("%X" % (case+baseoff))

        logger.info("OUTSTRINGS: %s (0x%x)", repr(outstrings), jmpva)

        idxs = '_'.join(outstrings)
        casename = "case_%s_%x" % (idxs, addr)

        curname = vw.getName(addr)
        if curname is not None:
            ## TODO:   if already labeled, chances are good this is other cases in the same function.
            ## either simply add the new outstrings to the current one or we need to keep track of what
            ## calls each and with what switchcase/index info.  VaSet?  or do we want this to only expect
            ## the same function to call each one, and all part of the same Switchcase?
            logger.info("%s is already labeled %s", casename, curname)

        vw.makeName(addr, casename)
        logger.info(casename)

    vw.makeName(jmpva, "switch_%.8x" % jmpva)
    logger.info("making switchname: switch_%.8x", jmpva)


def link_up(vw, jmpva, array, count, baseoff, baseva=None, itemsize=None):
    '''
    Manually link up a switchcase for which we have a pointer/offset array

    If baseva is None, the array is treated as a pointer array
    Otherwise, the numbers in the array are treated as offset from baseva

    If itemsize is not None, itemsize is treated as the size of each item
        in the array (in bytes. ie. 2, 4, 8 bytes)
    '''
    funcva = vw.getFunction(jmpva)
    lower = 0
    upper = count

    logger.info("link_up(0x%x, 0x%x, %d, 0x%x, %r, %r)", jmpva, array, count, baseoff, baseva, itemsize)

    cases = {}
    memrefs = []
    for idx in range(count):
        # handle specific itemsize if not pointer sized

        if itemsize is not None:
            idxloc = array + (idx * itemsize)
            fmt = e_bits.getFormat(itemsize, big_endian=vw.getEndian())
            addr, = vw.readMemoryFormat(idxloc, fmt)
        else:
            itemsize = vw.psize
            idxloc = array + (idx * itemsize)
            addr = vw.readMemoryPtr(idxloc)

        # sign-extend if an offset (many are negative)
        if itemsize < vw.psize:
            addr = e_bits.sign_extend(addr, itemsize, vw.psize)

        # handle base-va calculation
        if baseva is not None:
            addr = e_bits.unsigned(baseva + addr, vw.psize)

        logger.info("0x%x analyzeSwitch: idx: %s \t address: 0x%x", jmpva, idx, addr)

        # determining when to stop identifying switch-cases can be tough.  we assume that we have the 
        # correct number handed into this function in "count", but currently we'll stop analyzing
        # if we run into trouble.
        if not vw.isValidPointer(addr):
            logger.info("found invalid pointer.  quitting.  (0x%x in 0x%x)" % addr, jmpva)
            break

        tloc = vw.getLocation(addr)
        if tloc is not None and tloc[0] != addr:
            # pointing at something not right.  must be done.
            logger.info("found overlapping location.  quitting.")
            break

        if len(cases):
            xrefsto = vw.getXrefsTo(addr)
            # if there is an xref to this target from within this function, we're still ok... ish?
            if len(xrefsto):
                good = True
                for xrfr, xrto, xrt, xrtinfo in xrefsto:
                    xrfrfunc = vw.getFunction(xrfr)
                    if xrfrfunc in (funcva, None):
                        continue

                    # this one is *not* in the same function
                    logger.info("thisfunc: %r  otherfunc: %r", funcva, xrfrfunc)
                    #good = False

                if not good:
                    logger.info("target location (0x%x) has xrefs.", addr)
                    break

        # this is a valid thing, we have locations...  match them up
        caselist = cases.get(addr)
        if caselist is None:
            caselist = []
            cases[addr] = caselist
        caselist.append( idx )

        # make the connections
        vw.addXref(jmpva, addr, vivisect.REF_CODE)
        nloc = vw.getLocation(addr)
        if nloc is None:
            vw.makeCode(addr)

        # make pointers/numbers from the ptr/offset locations
        if baseva is not None:
            vw.makeNumber(idxloc, itemsize)
        else:
            vw.makePointer(idxloc)

    # makeNames (done separately because some indexes may be combined)
    makeNames(vw, jmpva, cases, baseoff)

    # store some metadata in a VaSet
    vw.setVaSetRow('SwitchCases', (jmpva, jmpva, count) )

    if baseoff is not None:
        lower += baseoff
        upper += baseoff

    vw.setComment(jmpva, "lower: 0x%x, upper: 0x%x" % (lower, upper))


def analyzeFunction(vw, fva, timeout=None):
    '''
    Function analysis module.
    This is inserted right after codeblock analysis
    '''
    targetNewFunctions(vw, fva)

    lastdynlen = 0
    dynbranches = vw.getVaSet('DynamicBranches')
    done = vw.getMeta('analyzedDynBranches')
    if done is None:
        done = []

    if not timeout:
        timeout = vw.config.viv.analysis.symswitchcase.timeout_secs

    # because the VaSet is often updated during analysis, we have to check to see if there are new 
    # dynamic branches to analyze.
    while lastdynlen != len(dynbranches):
        lastdynlen = len(dynbranches)
        for jmpva, (none, oprepr, bflags) in list(dynbranches.items()):
            if bflags & envi.BR_PROC:   # skip calls
                continue

            try:
                funcva = vw.getFunction(jmpva)
                if funcva != fva:
                    # jmp_indir is for the entire VivWorkspace.  
                    # we're just filtering to this function here.
                    # this should be checked again when codeblocks are allowed to 
                    #   be part of multiple functions.
                    continue

                if jmpva in done:
                    logger.info("...skipping 0x%x - already done", jmpva)
                    continue

                done.append(jmpva)
                analyzeJmp(vw, jmpva, timeout)
            except:
                logger.info('Exception processing SwitchCase in function 0x%x', fva, exc_info=1)

        dynbranches = vw.getVaSet('DynamicBranches')
    vw.setMeta('analyzedDynBranches', done)
    # TODO: we need a better way to store changing lists/dicts, that don't show up in the UI.  VaSet would be great, but ugly

def analyzeJmp(vw, jmpva, timeout=None):
    if vw.getVaSetRow('SwitchCases', jmpva) is not None:
        logger.info("...skipping 0x%x - SwitchCases already has it?", jmpva)
        return

    if vw.getVaSetRow('SwitchCases_TimedOut', jmpva) is not None:
        logger.warning("Reanalyzing TimedOut SwitchCase: 0x%x (timeout=%r secs)", jmpva, timeout)
        logger.debug("Removing 0x%x from VaSet SwitchCases_TimedOut", jmpva)
        vw.delVaSetRow('SwitchCases_TimedOut', jmpva)

    try:
        sc = SwitchCase(vw, jmpva, timeout)
        sc.analyze()
    except e_exc.ArchNotImplemented:
        logger.info("... skipping.  Current architecture has no SymbolikTranslator")






import unittest

import cobra.devent as c_devent

class CobraEventTest(unittest.TestCase):

    #def setUp(self):
        #pass

    #def tearDown(self):
        #pass

    def test_cobra_devent(self):

        eventcore = c_devent.CobraEventCore()
        chan = eventcore.initEventChannel()
        eventcore.fireEvent('blah',(1,2,3))
        etup = eventcore.getNextEventsForChan(chan)[0]
        self.assertEqual(etup[0], 'blah')

    def test_cobra_devent_timeout(self):
        eventcore = c_devent.CobraEventCore()
        chan = eventcore.initEventChannel()
        self.assertFalse( eventcore.getNextEventsForChan( chan, timeout=0.01 ) )



import os
import sys
import time
import tempfile
import unittest
import multiprocessing as mp

import vivisect.cli as vivcli
import vivisect.const as v_const
import vivisect.tests.helpers as helpers
import vivisect.remote.server as v_r_server

import envi.memcanvas as e_mcanvas
import envi.common as ecmn
logger = ecmn.logging.getLogger(__name__)
ecmn.initLogging(logger)

def runServer(name, port):
    dirn = os.path.dirname(name)
    testfile = helpers.getTestPath('windows', 'amd64', 'firefox.exe')

    # load the file in so we get some workspace events, but not so many to make
    # this test take forever
    vw = vivcli.VivCli()
    vw.loadFromFile(testfile)
    vw.setMeta('StorageName', name)
    vw.saveWorkspace()

    v_r_server.runMainServer(dirn, port)


def getOutput(strcanv, timeout=10):
    '''
    Watches a StringCanvas for data up to a given timeout
    '''
    starttime = time.time()
    output = strcanv.strval
    while not output and (time.time() - starttime) < timeout:
        time.sleep(.1)
        output = strcanv.strval

    strcanv.strval = ''
    return output

class VivisectRemoteTests(unittest.TestCase):
    '''
    So...what would be fun is basically a chain of remote workspaces all tied in interesting
    configurations.
    '''
    def test_basic(self):
        testfile = helpers.getTestPath('windows', 'amd64', 'firefox.exe')
        good = vivcli.VivCli()
        good.loadFromFile(testfile)

        host = 'localhost'
        port = 0x4097
        with tempfile.TemporaryDirectory() as tmpd:
            tmpf = tempfile.NamedTemporaryFile(dir=tmpd, delete=False)
            try:
                proc = mp.Process(target=runServer, args=(tmpf.name, port,))
                proc.daemon = True
                proc.start()

                # give the other process time to spin up
                time.sleep(0.5)

                # So...yea. The server could not be up yet, but I'm not waiting a million years to
                # wait for it.
                retry = 0
                conn = False
                while retry < 5:
                    try:
                        server = v_r_server.connectToServer(host, port)
                        conn = True
                        break
                    except:
                        retry += 1
                        time.sleep(0.2)

                if not conn:
                    self.fail('Could not connect to %s:%s' % (host, port))

                wslist = server.listWorkspaces()
                self.assertEqual(len(wslist), 1)
                self.assertEqual(server.getServerVersion(), 20130820)

                rmtvw = v_r_server.getServerWorkspace(server, wslist[0])
                rmtvw2 = v_r_server.getServerWorkspace(server, wslist[0])
                # So the consumption of events from the server is *also* threaded, so I've got to do some blocking
                # to get us to wait on things
                retry = 0
                while retry < 5:
                    locs = rmtvw2.getLocations()
                    if len(locs) > 1388:
                        break

                    retry += 1
                    time.sleep(0.3)
                    sys.stderr.write('%d' % retry)

                self.assertEqual(len(rmtvw2.getLocations()), 1389)
                self.assertEqual(set(rmtvw2.getLocations()), set(good.getLocations()))
                self.assertEqual(set(rmtvw2.getXrefs()), set(good.getXrefs()))

                # test some of the follow-the-leader framework
                testuuid = 'uuid_of_my_dearest_friend_1'
                # first just create a leader session:
                rmtvw.iAmLeader(testuuid, "atlas' moving castle")

                retry = 0
                while retry < 5:
                    # only one session, so we'll run this once - local
                    ldrsess = rmtvw2.getLeaderSessions().get(testuuid)
                    if ldrsess is not None:
                        break

                    retry += 1
                    time.sleep(.1)
                    sys.stderr.write('%d' % retry)

                (user, fname) = ldrsess
                self.assertEqual(fname, "atlas' moving castle")

                # now let's move around a bit
                rmtvw.followTheLeader(testuuid, '0x31337')
                retry = 0
                while retry < 5:
                    # only one session, so we'll run this once - local
                    ldrloc = rmtvw2.getLeaderLoc(testuuid)
                    if ldrloc is not None:
                        break

                    retry += 1
                    time.sleep(.1)
                    sys.stderr.write('%d' % retry)

                self.assertEqual(ldrloc, '0x31337')

                # now let's rename things
                rmtvw.modifyLeaderSession(testuuid, 'rakuy0', "rakuy0's moving castle")
                retry = 0
                while retry < 5:
                    # only one session, so we'll run this once - local
                    ldrsess = list(rmtvw2.getLeaderSessions().items())[0]
                    uuid, (user, fname) = ldrsess
                    if user == 'rakuy0':
                        break

                    retry += 1
                    time.sleep(.1)
                    sys.stderr.write('%d' % retry)

                self.assertEqual(uuid, testuuid)
                self.assertEqual(user, 'rakuy0')
                self.assertEqual(fname, "rakuy0's moving castle")

                self.assertEqual(rmtvw2.getLeaderInfo(testuuid), ('rakuy0', "rakuy0's moving castle"))


                # test the CLI:
                rmtvw.canvas = e_mcanvas.StringMemoryCanvas(rmtvw)

                # empty leaders
                rmtvw.do_leaders('')
                output = getOutput(rmtvw.canvas)
                self.assertIn('Manage Leader Sessions.', output)

                # list leaders
                rmtvw.do_leaders('list')
                output = getOutput(rmtvw.canvas)
                self.assertIn("rakuy0's moving castle", output)

                # modify a leader session
                rmtvw.do_leaders('mod %s foo foo bar baz' % testuuid)
                retry = 0
                while retry < 5:
                    ldrsess = rmtvw2.getLeaderSessions().get(testuuid)
                    if ldrsess:
                        user, fname = ldrsess
                        if user == 'foo':
                            break

                    retry += 1
                    time.sleep(.1)
                    sys.stderr.write('%d' % retry)

                output = getOutput(rmtvw.canvas)
                self.assertIn("foo bar baz", output)

                # kill leader session
                rmtvw.do_leaders('kill %s' % testuuid)
                retry = 0
                while retry < 5:
                    ldrsess = rmtvw2.getLeaderSessions().get(testuuid)
                    if not ldrsess:
                        break

                    retry += 1
                    time.sleep(.1)
                    sys.stderr.write('%d' % retry)

                output = getOutput(rmtvw.canvas)
                self.assertIn("*Ended: foo's session 'foo bar baz'", output)


                # kill all (which means we need to add a few)
                #   repopulate
                rmtvw.iAmLeader(testuuid+'1', "castle 1")
                rmtvw.iAmLeader(testuuid+'2', "castle 2")
                rmtvw.iAmLeader(testuuid+'3', "castle 3")

                retry = 0
                while retry < 5:
                    # only one session, so we'll run this once - local
                    ldrsess = rmtvw2.getLeaderSessions().get(testuuid+'3')
                    if ldrsess is not None:
                        break

                    retry += 1
                    time.sleep(.1)
                    sys.stderr.write('%d' % retry)

                #   kill them all!
                rmtvw.do_leaders('killall')
                retry = 0
                while retry < 5:
                    ldrsessions = rmtvw2.getLeaderSessions()
                    if not len(ldrsessions):
                        break

                    retry += 1
                    time.sleep(.1)
                    sys.stderr.write('%d' % retry)

                output = getOutput(rmtvw.canvas)
                self.assertIn("castle 3", output)
                self.assertEqual(len(rmtvw2.getLeaderSessions()), 0)

                # close down tests
                try:
                    rmtvw.server = None
                    rmtvw2.server = None

                    q = rmtvw.chan_lookup.get(rmtvw.rchan)
                    if q:
                        # So it's not reeeealy auto analysis fini, but it's a good enough stand-in to get
                        # the server thread to shutdown cleaner
                        q.put((v_const.VWE_AUTOANALFIN, None))

                except:
                    pass
            finally:
                try:
                    proc.terminate()
                    proc.join(timeout=5)
                except:
                    pass
                try:
                    proc.close()
                except:
                    pass
                tmpf.close()
                if os.path.exists(tmpf.name):
                    os.unlink(tmpf.name)



'''
Quick utility to generate ord lookups from DLL exports.
'''

import sys
import argparse

import PE


def setup():
    ap = argparse.ArgumentParser('Quick Utility to generate ordlookups from DLLs')
    ap.add_argument('file', help='Path to PE file')
    return ap


def main(argv):
    opts = setup().parse_args(argv)
    with open(opts.file, 'rb') as f:
        p = PE.PE(f)

        base = p.IMAGE_EXPORT_DIRECTORY.Base

        ords = {}
        for fva, ord, name in p.getExports():
            ords[ord+base] = name

        keys = list(ords.keys())
        for k in keys.sort():
            print('    %d:"%s",' % (k, ords.get(k)))


if __name__ == '__main__':
    sys.exit(main(sys.argv[1:]))
